(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("base" "draw-lib" "gui-lib" "snip-lib" "htdp-lib"))) (define build-deps (quote ("racket-doc" "htdp-doc" "scribble-lib"))) (define pkg-desc "Teaching libraries for _Picturing Programs_") (define pkg-authors (quote (sbloch))) (define license (quote (Apache-2.0 OR MIT)))))
