(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("srfi-lib" ("srfi-doc" #:version "1.1")))) (define implies (quote ("srfi-lib" "srfi-doc"))) (define pkg-desc "Legacy SRFI (Scheme) libraries") (define pkg-authors (quote (mflatt noel chongkai jay))) (define license (quote (Apache-2.0 OR MIT)))))
