(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("scheme-lib" "typed-racket-lib" ("base" #:version "8.5.0.3")))) (define pkg-desc "compatibility library for older Typed Racket-based languages") (define pkg-authors (quote (samth stamourv))) (define version "1.8") (define license (quote (Apache-2.0 OR MIT)))))
