#lang racket/base

(require rackunit/docs-complete)

#;(check-docs (quote frtime/frtime-big))
#;(check-docs (quote frtime/gui/fred))
#;(check-docs (quote frtime/gui))
#;(check-docs (quote frtime/animation))
#;(check-docs (quote frtime/frlibs/math))
