(module constants mzscheme
  (provide machenrys-constant)
  
  ;; machenrys-constant nonnegative?
  ;; the differents between the size a pasteboard is alloted by
  ;; get-view-size and the size it needs to avoid scollbars
  (define machenrys-constant 2))
