(module info setup/infotab (#%module-begin (define collection "shell-completion") (define deps (quote ("base"))) (define pkg-desc "Completion scribpts for bash and zsh") (define pkg-authors (quote (samth sstrickl stamourv))) (define test-responsibles (quote (("racket-completion.zsh" eli)))) (define license (quote (Apache-2.0 OR MIT)))))
