#lang info

;; the XML tool has been moved to the stepper collection, so that the
;; stepper can create xml snips.  See collects/stepper/tool.rkt for (a
;; bit) more information
(define tools '(("text-box-tool.rkt")))
(define tool-names '("Text Box"))
