(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("redex-doc" "redex-examples" "redex-lib" "redex-gui-lib"))) (define implies (quote ("redex-doc" "redex-examples" "redex-lib" "redex-gui-lib"))) (define pkg-desc "PLT Redex libraries for practical semantics engineering") (define pkg-authors (quote (robby bfetscher))) (define license (quote (Apache-2.0 OR MIT)))))
