#lang s-exp syntax/module-reader

typed/racket/shallow

#:read r:read
#:read-syntax r:read-syntax
#:info make-info
#:language-info make-language-info

(define (make-info key default use-default)
  (case key
    [(drracket:opt-in-toolbar-buttons)
     '(optimization-coach)]
    [else (use-default key default)]))

(define make-language-info
  `#(typed-racket/language-info get-info (shallow)))


(require (prefix-in r: typed-racket/typed-reader))
