(module info setup/infotab (#%module-begin (define collection (quote multi)) (define build-deps (quote ("scheme-lib" "draw-doc" "gui-doc" "pict-doc" "scribble-doc" "base" "gui-lib" "pict-lib" "scribble-lib" "slideshow-lib" ("racket-doc" #:version "1.3") "at-exp-lib"))) (define update-implies (quote ("slideshow-lib"))) (define pkg-desc "documentation part of \"slideshow\"") (define pkg-authors (quote (mflatt robby))) (define license (quote (Apache-2.0 OR MIT)))))
