(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("base"))) (define build-deps (quote ("scheme-lib" "racket-doc" "syntax-color-doc" "parser-tools-lib" "scribble-lib"))) (define update-implies (quote ("parser-tools-lib"))) (define pkg-desc "documentation part of \"parser-tools\"") (define pkg-authors (quote (mflatt))) (define license (quote (Apache-2.0 OR MIT)))))
