(module info setup/infotab (#%module-begin (define collection "expeditor") (define deps (quote (("base" #:version "8.3.0.11") ("syntax-color-lib" #:version "1.3")))) (define pkg-desc "implementation (no documentation) part of \"expeditor\"") (define pkg-authors (quote (mflatt))) (define version "1.1") (define license (quote (Apache-2.0 OR MIT)))))
