#lang info

(define scribblings
  '(("start.scrbl" (user-doc-root depends-all no-depend-on) (omit))
    ("search.scrbl" (user-doc depends-all-user no-depend-on) (omit))
    ("local-redirect.scrbl" (user-doc depends-all-user no-depend-on) (omit))
    ("release.scrbl" (user-doc depends-all no-depend-on) (omit))))
