#lang racket/base
(require rackunit/docs-complete)
(check-docs (quote teachpack/world))
(check-docs (quote teachpack/value-turtles))
(check-docs (quote teachpack/turtles))
(check-docs (quote teachpack/server))
(check-docs (quote teachpack/picturing-programs))
(check-docs (quote teachpack/nuworld))
(check-docs (quote teachpack/door))
(check-docs (quote teachpack/balls))
