#lang info

(define test-omit-paths '("xmmsctrl.rkt"
                          "use-xmmsctrl.rkt"
                          "esd.rkt"
                          "magick.rkt"
                          "sndfile.rkt"
                          "tcl.rkt"
                          "use-esd.rkt"
                          "use-magick.rkt"
                          "use-sndfile.rkt"
                          "use-tcl.rkt"
                          "use-xosd.rkt"
                          "xosd.rkt"))

(define test-responsibles '(("sndfile.rkt" clements)
                            (all eli)))
