(module info setup/infotab (#%module-begin (define deps (quote ("base" "drracket-plugin-lib" "drracket" "gui-lib" "net-lib" "scribble-lib"))) (define build-deps (quote ("at-exp-lib" "drracket" "gui-doc" "racket-doc" "draw-doc" "rackunit-lib"))) (define name "Quickscript") (define drracket-tools (quote (("tool.rkt")))) (define drracket-tool-names (quote ("Quickscript"))) (define drracket-tool-icons (quote (#f))) (define scribblings (quote (("scribblings/quickscript.scrbl" () (drracket-plugin) "quickscript")))) (define compile-omit-paths (quote ())) (define license (quote (Apache-2.0 OR MIT)))))
