#lang scheme/signature

hierarchical-list%
hierarchical-list-item<%>
hierarchical-list-item%
hierarchical-list-compound-item<%>
hierarchical-list-compound-item%

hierarchical-item-snip%
hierarchical-list-snip%
