(module info setup/infotab (#%module-begin (define collection "profile") (define deps (quote ("base" "errortrace-lib"))) (define build-deps (quote ("at-exp-lib" "rackunit-lib"))) (define pkg-desc "implementation (no documentation) part of \"profile\"") (define pkg-authors (quote (eli stamourv))) (define raco-commands (quote (("profile" profile/raco "profile execution time" #f)))) (define version "1.1") (define license (quote (Apache-2.0 OR MIT)))))
