(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("typed-racket-lib" "typed-racket-doc"))) (define implies (quote ("typed-racket-lib" "typed-racket-doc"))) (define pkg-desc "The implementation of the Typed Racket language") (define pkg-authors (quote (samth stamourv))) (define version "1.15") (define license (quote (Apache-2.0 OR MIT)))))
