(module info setup/infotab (#%module-begin (define collection "lazy") (define scribblings (quote (("lazy.scrbl" () (experimental 50))))) (define drracket-tools (quote (("lazy-tool.rkt")))) (define drracket-tool-names (quote ("Lazy Racket"))) (define deps (quote ("base" "drracket-plugin-lib" "htdp-lib" "string-constants-lib" "compatibility-lib"))) (define build-deps (quote ("mzscheme-doc" "scheme-lib" "eli-tester" "racket-doc" "scribble-lib"))) (define pkg-desc "The implementation of the Lazy Racket language") (define pkg-authors (quote (eli stchang))) (define license (quote (Apache-2.0 OR MIT)))))
