(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("syntax-color-lib" "syntax-color-doc"))) (define implies (quote ("syntax-color-lib" "syntax-color-doc"))) (define pkg-desc "Program syntax coloring for editors and typesetting") (define pkg-authors (quote (mflatt))) (define license (quote (Apache-2.0 OR MIT)))))
