(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("simple-tree-text-markup-lib" "simple-tree-text-markup-doc"))) (define implies (quote ("simple-tree-text-markup-lib" "simple-tree-text-markup-doc"))) (define pkg-desc "Combinator library for simple markup, mainly for displaying messages in a REPL") (define pkg-authors (quote (sperber))) (define license (quote (Apache-2.0 OR MIT)))))
