(module info setup/infotab (#%module-begin (define deps (quote ("base"))) (define collection "mrlib") (define pkg-desc "Table of TeX-style abbreviations") (define pkg-authors (quote (robby))) (define license (quote (Apache-2.0 OR MIT)))))
