#lang setup/infotab

(define scribblings '(("scribblings/optimization-coach.scrbl" () ("Performance Tools"))))

(define drracket-tools '(("tool.rkt")))
(define drracket-tool-names '("Optimization Coach"))
